@@ -28,6 +28,7 @@
 #include "fdeep/common.hpp"
 
 #include "fdeep/layers/add_layer.hpp"
+#include "fdeep/layers/additive_attention_layer.hpp"
 #include "fdeep/layers/attention_layer.hpp"
 #include "fdeep/layers/average_layer.hpp"
 #include "fdeep/layers/average_pooling_3d_layer.hpp"
@@ -50,6 +51,8 @@
 #include "fdeep/layers/hard_sigmoid_layer.hpp"
 #include "fdeep/layers/input_layer.hpp"
 #include "fdeep/layers/layer.hpp"
+#include "fdeep/layers/layer_normalization_layer.hpp"
+#include "fdeep/layers/unit_normalization_layer.hpp"
 #include "fdeep/layers/leaky_relu_layer.hpp"
 #include "fdeep/layers/embedding_layer.hpp"
 #include "fdeep/layers/lstm_layer.hpp"
@@ -61,6 +64,7 @@
 #include "fdeep/layers/maximum_layer.hpp"
 #include "fdeep/layers/minimum_layer.hpp"
 #include "fdeep/layers/model_layer.hpp"
+#include "fdeep/layers/multi_head_attention_layer.hpp"
 #include "fdeep/layers/multiply_layer.hpp"
 #include "fdeep/layers/normalization_layer.hpp"
 #include "fdeep/layers/pooling_3d_layer.hpp"
@@ -535,6 +539,29 @@ inline layer_ptr create_batch_normalization_layer(const get_param_f& get_param,
         name, axis, moving_mean, moving_variance, beta, gamma, epsilon);
 }
 
+
+inline layer_ptr create_layer_normalization_layer(const get_param_f& get_param,
+    const nlohmann::json& data, const std::string& name)
+{
+    const bool center = data["config"]["center"];
+    const bool scale = data["config"]["scale"];
+    const auto axes = create_vector<int>(create_int, data["config"]["axis"]);
+    const float_type epsilon = data["config"]["epsilon"];
+    float_vec gamma;
+    float_vec beta;
+    if (scale) gamma = decode_floats(get_param(name, "gamma"));
+    if (center) beta = decode_floats(get_param(name, "beta"));
+    return std::make_shared<layer_normalization_layer>(
+        name, axes, beta, gamma, epsilon);
+}
+
+inline layer_ptr create_unit_normalization_layer(const get_param_f&,
+    const nlohmann::json& data, const std::string& name)
+{
+    const auto axes = create_vector<int>(create_int, data["config"]["axis"]);
+    return std::make_shared<unit_normalization_layer>(name, axes);
+}
+
 inline layer_ptr create_identity_layer(
     const get_param_f&, const nlohmann::json&, const std::string& name)
 {
@@ -621,7 +648,7 @@ inline layer_ptr create_concatenate_layer(
     const get_param_f&, const nlohmann::json& data,
     const std::string& name)
 {
-    const std::int32_t keras_axis = data["config"]["axis"];
+    const int keras_axis = data["config"]["axis"];
     return std::make_shared<concatenate_layer>(name, keras_axis);
 }
 
@@ -650,7 +677,7 @@ inline layer_ptr create_dot_layer(
     const get_param_f&, const nlohmann::json& data,
     const std::string& name)
 {
-    const auto axes = create_vector<std::size_t>(create_size_t, data["config"]["axes"]);
+    const auto axes = create_vector<int>(create_int, data["config"]["axes"]);
     const bool normalize = data["config"]["normalize"];
     return std::make_shared<dot_layer>(name, axes, normalize);
 }
@@ -1014,12 +1041,56 @@ inline layer_ptr create_category_encoding_layer(
 }
 
 inline layer_ptr create_attention_layer(
-    const get_param_f&,
+    const get_param_f& get_param,
     const nlohmann::json& data, const std::string& name)
 {
     const bool use_scale = data["config"]["use_scale"];
     const std::string score_mode = data["config"]["score_mode"];
-    return std::make_shared<attention_layer>(name, use_scale, score_mode);
+    float_type scale = static_cast<float_type>(1);
+    float_type concat_score_weight = static_cast<float_type>(1);
+    if (use_scale) {
+        scale = get_param(name, "scale");
+    }
+    if (score_mode == "concat") {
+        concat_score_weight = get_param(name, "concat_score_weight");
+    }
+    return std::make_shared<attention_layer>(name, score_mode, scale, concat_score_weight);
+}
+
+inline layer_ptr create_additive_attention_layer(
+    const get_param_f& get_param,
+    const nlohmann::json& data, const std::string& name)
+{
+    const bool use_scale = data["config"]["use_scale"];
+    float_vec scale(static_cast<float_type>(1), 1);
+    if (use_scale) {
+        scale = decode_floats(get_param(name, "scale"));
+    }
+    return std::make_shared<additive_attention_layer>(name, scale);
+}
+
+inline layer_ptr create_multi_head_attention_layer(
+    const get_param_f& get_param,
+    const nlohmann::json& data, const std::string& name)
+{
+    const std::size_t num_heads = data["config"]["num_heads"];
+    const std::size_t key_dim = data["config"]["key_dim"];
+    const std::size_t value_dim = data["config"]["value_dim"];
+    const bool use_bias = data["config"]["use_bias"];
+    const auto weight_shapes =
+        create_vector<std::vector<std::size_t>>(fplus::bind_1st_of_2(
+            create_vector<std::size_t, decltype(create_size_t)>, create_size_t),
+            get_param(name, "weight_shapes"));
+    const auto weight_values = create_vector<float_vec>(decode_floats, get_param(name, "weights"));
+    const auto weights_and_biases = fplus::zip_with(
+        [](const std::vector<std::size_t>& shape, const float_vec& values) -> tensor
+        {
+            return tensor(
+                create_tensor_shape_from_dims(shape),
+                fplus::convert_container<float_vec>(values));
+        }, weight_shapes, weight_values);
+    return std::make_shared<multi_head_attention_layer>(name,
+        num_heads, key_dim, value_dim, use_bias, weights_and_biases);
 }
 
 inline std::string get_activation_type(const nlohmann::json& data)
@@ -1095,11 +1166,35 @@ inline node create_node(const nlohmann::json& inbound_nodes_data)
             inbound_nodes_data));
 }
 
+inline nodes create_multi_head_attention_nodes(const std::vector<nlohmann::json> inbound_nodes_data)
+{
+    assertion(inbound_nodes_data.size() == 1 && inbound_nodes_data.front().size() == 1,
+        "multi_head_attention needs to have exactly one primary inbound node; see https://stackoverflow.com/q/77400589/1866775");
+    const auto inbound_node_data = inbound_nodes_data.front().front();
+    const auto value = inbound_node_data[3]["value"];
+    if (json_obj_has_member(inbound_node_data[3], "key")) {
+        return {
+            node({
+                create_node_connection(inbound_node_data),
+                create_node_connection(value),
+                create_node_connection(inbound_node_data[3]["key"])
+                })};
+    }
+    return {
+        node({
+            create_node_connection(inbound_node_data),
+            create_node_connection(value)
+            })};
+}
+
 inline nodes create_nodes(const nlohmann::json& data)
 {
     assertion(data["inbound_nodes"].is_array(), "no inbound nodes");
     const std::vector<nlohmann::json> inbound_nodes_data =
         data["inbound_nodes"];
+    if (data["class_name"] == "MultiHeadAttention") {
+        return create_multi_head_attention_nodes(inbound_nodes_data);
+    }
     return fplus::transform(create_node, inbound_nodes_data);
 }
 
@@ -1262,6 +1357,8 @@ inline layer_ptr create_layer(const get_param_f& get_param,
             {"DepthwiseConv2D", create_depthwise_conv_2D_layer},
             {"InputLayer", create_input_layer},
             {"BatchNormalization", create_batch_normalization_layer},
+            {"LayerNormalization", create_layer_normalization_layer},
+            {"UnitNormalization", create_unit_normalization_layer},
             {"Dropout", create_identity_layer},
             {"ActivityRegularization", create_identity_layer},
             {"AlphaDropout", create_identity_layer},
@@ -1329,6 +1426,8 @@ inline layer_ptr create_layer(const get_param_f& get_param,
             {"Normalization", create_normalization_layer},
             {"CategoryEncoding", create_category_encoding_layer},
             {"Attention", create_attention_layer},
+            {"AdditiveAttention", create_additive_attention_layer},
+            {"MultiHeadAttention", create_multi_head_attention_layer},
         };
 
     const wrapper_layer_creators wrapper_creators = {
