@@ -325,6 +325,18 @@ def show_batch_normalization_layer(layer):
     return result
 
 
+def show_layer_normalization_layer(layer):
+    """Serialize layer normalization layer to dict"""
+    result = {}
+    if layer.center:
+        beta = K.get_value(layer.beta)
+        result['beta'] = encode_floats(beta)
+    if layer.scale:
+        gamma = K.get_value(layer.gamma)
+        result['gamma'] = encode_floats(gamma)
+    return result
+
+
 def show_dense_layer(layer):
     """Serialize dense layer to dict"""
     weights = layer.get_weights()
@@ -405,7 +417,7 @@ def show_gru_layer(layer):
 
 def transform_cudnn_weights(input_weights, recurrent_weights, n_gates):
     return transform_kernels(input_weights, n_gates, transform_input_kernel), \
-           transform_kernels(recurrent_weights, n_gates, transform_recurrent_kernel)
+        transform_kernels(recurrent_weights, n_gates, transform_recurrent_kernel)
 
 
 def show_cudnn_lstm_layer(layer):
@@ -518,6 +530,11 @@ def show_resizing_layer(layer):
     assert layer.interpolation in ["nearest", "bilinear", "area"]
 
 
+def show_rescaling_layer(layer):
+    """Serialize Rescaling layer to dict"""
+    assert isinstance(layer.scale, float)
+
+
 def show_category_encoding_layer(layer):
     """Serialize CategoryEncoding layer to dict"""
     assert layer.output_mode in ["multi_hot", "count", "one_hot"]
@@ -525,8 +542,35 @@ def show_category_encoding_layer(layer):
 
 def show_attention_layer(layer):
     """Serialize Attention layer to dict"""
-    assert layer.score_mode in ["dot"]
-    assert layer.use_scale is False
+    assert layer.score_mode in ["dot", "concat"]
+    data = {}
+    if layer.scale:
+        data['scale'] = float(layer.scale.numpy())
+    if layer.score_mode == "concat":
+        data['concat_score_weight'] = float(layer.concat_score_weight.numpy())
+    if data:
+        return data
+
+
+def show_additive_attention_layer(layer):
+    """Serialize AdditiveAttention layer to dict"""
+    data = {}
+    if layer.scale is not None:
+        data['scale'] = encode_floats(layer.scale.numpy())
+    if data:
+        return data
+
+
+def show_multi_head_attention_layer(layer):
+    """Serialize MultiHeadAttention layer to dict"""
+    assert len(layer.input_shape) == 3
+    assert layer.input_shape[0] is None
+    assert layer._output_shape is None
+    assert layer._attention_axes == (1,), "MultiHeadAttention supported only with attention_axes=None"
+    return {
+        'weight_shapes': list(map(lambda w: list(w.shape), layer.weights)),
+        'weights': list(map(lambda w: encode_floats(w.numpy()), layer.weights)),
+    }
 
 
 def get_layer_functions_dict():
@@ -540,6 +584,7 @@ def get_layer_functions_dict():
         'Dot': show_dot_layer,
         'PReLU': show_prelu_layer,
         'Embedding': show_embedding_layer,
+        'LayerNormalization': show_layer_normalization_layer,
         'LSTM': show_lstm_layer,
         'GRU': show_gru_layer,
         'CuDNNLSTM': show_cudnn_lstm_layer,
@@ -551,8 +596,11 @@ def get_layer_functions_dict():
         'Normalization': show_normalization_layer,
         'UpSampling2D': show_upsampling2d_layer,
         'Resizing': show_resizing_layer,
+        'Rescaling': show_rescaling_layer,
         'CategoryEncoding': show_category_encoding_layer,
         'Attention': show_attention_layer,
+        'AdditiveAttention': show_additive_attention_layer,
+        'MultiHeadAttention': show_multi_head_attention_layer,
     }
 
 
@@ -846,7 +894,7 @@ def convert(in_path, out_path, no_tests=False):
     workaround_cudnn_not_found_problem()
 
     print('loading {}'.format(in_path))
-    model = load_model(in_path)
+    model = load_model(in_path, compile=False)
     json_output = model_to_fdeep_json(model, no_tests)
     print('writing {}'.format(out_path))
 
@@ -859,7 +907,7 @@ def main():
 
     parser = argparse.ArgumentParser(
         prog='frugally-deep model converter',
-        description='Converts models from Keras\' .h5 format to frugally-deep\'s .json format.')
+        description='Converts models from Keras\' .keras format to frugally-deep\'s .json format.')
     parser.add_argument('input_path', type=str)
     parser.add_argument('output_path', type=str)
     parser.add_argument('--no-tests', action='store_true')
