@@ -6,7 +6,9 @@ import sys
 
 import numpy as np
 from tensorflow.keras.layers import ActivityRegularization
-from tensorflow.keras.layers import BatchNormalization, Concatenate
+from tensorflow.keras.layers import AdditiveAttention
+from tensorflow.keras.layers import Attention
+from tensorflow.keras.layers import BatchNormalization, Concatenate, LayerNormalization, UnitNormalization
 from tensorflow.keras.layers import Bidirectional, TimeDistributed
 from tensorflow.keras.layers import CategoryEncoding
 from tensorflow.keras.layers import Conv1D, ZeroPadding1D, Cropping1D
@@ -21,11 +23,11 @@ from tensorflow.keras.layers import LeakyReLU, ELU, PReLU, ReLU
 from tensorflow.keras.layers import MaxPooling1D, AveragePooling1D, UpSampling1D
 from tensorflow.keras.layers import MaxPooling2D, AveragePooling2D, UpSampling2D
 from tensorflow.keras.layers import MaxPooling3D, AveragePooling3D
+from tensorflow.keras.layers import MultiHeadAttention
 from tensorflow.keras.layers import Multiply, Add, Subtract, Average, Maximum, Minimum, Dot
 from tensorflow.keras.layers import Permute, Reshape, RepeatVector
 from tensorflow.keras.layers import SeparableConv2D, DepthwiseConv2D
 from tensorflow.keras.layers import ZeroPadding3D, Cropping3D
-from tensorflow.keras.layers import Attention
 from tensorflow.keras.models import Model, load_model, Sequential
 
 __author__ = "Tobias Hermann"
@@ -140,7 +142,7 @@ def get_test_model_exhaustive():
         (2, 1, 1, 1, 1),
         (1, 1, 4, 1, 6),
         (1, 3, 1, 5, 1),
-        (2, 1, 4, 1, 1), # 45
+        (2, 1, 4, 1, 1),  # 45
         (1,),
         (3, 1),
         (6, 5, 4, 3, 2),
@@ -233,12 +235,19 @@ def get_test_model_exhaustive():
     outputs.append(Permute((2, 1, 3))(inputs[4]))
     outputs.append(Permute((2, 1))(inputs[6]))
     outputs.append(Permute((1,))(inputs[8]))
-
     outputs.append(Permute((3, 1, 2))(inputs[31]))
     outputs.append(Permute((3, 1, 2))(inputs[32]))
-    outputs.append(BatchNormalization()(Permute((3, 1, 2))(inputs[31])))
-    outputs.append(BatchNormalization()(Permute((3, 1, 2))(inputs[32])))
 
+    outputs.append(BatchNormalization(center=False, scale=False)(inputs[11]))
+    outputs.append(BatchNormalization()(inputs[11]))
+    outputs.append(BatchNormalization()(inputs[10]))
+    outputs.append(BatchNormalization()(inputs[14]))
+    outputs.append(BatchNormalization()(inputs[26]))
+    outputs.append(BatchNormalization()(inputs[23]))
+    outputs.append(BatchNormalization()(inputs[0]))
+    outputs.append(BatchNormalization(center=False)(inputs[0]))
+    outputs.append(BatchNormalization(scale=False)(inputs[0]))
+    outputs.append(BatchNormalization(center=False, scale=False)(inputs[0]))
     outputs.append(BatchNormalization()(inputs[0]))
     outputs.append(BatchNormalization(axis=1)(inputs[0]))
     outputs.append(BatchNormalization(axis=2)(inputs[0]))
@@ -282,6 +291,32 @@ def get_test_model_exhaustive():
     outputs.append(BatchNormalization(axis=4)(inputs[20]))
     outputs.append(BatchNormalization(axis=5)(inputs[20]))
 
+    outputs.append(LayerNormalization()(inputs[11]))
+    outputs.append(LayerNormalization()(inputs[10]))
+    outputs.append(LayerNormalization()(inputs[26]))
+    outputs.append(LayerNormalization()(inputs[24]))
+    outputs.append(LayerNormalization()(inputs[0]))
+    outputs.append(LayerNormalization(axis=1)(inputs[0]))
+    outputs.append(LayerNormalization(axis=2)(inputs[0]))
+    outputs.append(LayerNormalization(axis=3)(inputs[0]))
+    outputs.append(LayerNormalization(axis=4)(inputs[0]))
+    outputs.append(LayerNormalization(axis=5)(inputs[0]))
+    outputs.append(LayerNormalization(axis=[1, 2])(inputs[0]))
+    outputs.append(LayerNormalization(axis=[2, 3, 5])(inputs[0]))
+
+    outputs.append(UnitNormalization()(inputs[11]))
+    outputs.append(UnitNormalization()(inputs[10]))
+    outputs.append(UnitNormalization()(inputs[26]))
+    outputs.append(UnitNormalization()(inputs[24]))
+    outputs.append(UnitNormalization()(inputs[0]))
+    outputs.append(UnitNormalization(axis=1)(inputs[0]))
+    outputs.append(UnitNormalization(axis=2)(inputs[0]))
+    outputs.append(UnitNormalization(axis=3)(inputs[0]))
+    outputs.append(UnitNormalization(axis=4)(inputs[0]))
+    outputs.append(UnitNormalization(axis=5)(inputs[0]))
+    outputs.append(UnitNormalization(axis=[1, 2])(inputs[0]))
+    outputs.append(UnitNormalization(axis=[2, 3, 5])(inputs[0]))
+
     outputs.append(Dropout(0.5)(inputs[4]))
     outputs.append(ActivityRegularization(0.3, 0.4)(inputs[4]))
 
@@ -393,6 +428,47 @@ def get_test_model_exhaustive():
 
     outputs.append(Attention(use_scale=False, score_mode='dot')([inputs[49], inputs[50]]))
     outputs.append(Attention(use_scale=False, score_mode='dot')([inputs[49], inputs[50], inputs[51]]))
+    outputs.append(Attention(use_scale=True, score_mode='dot')([inputs[49], inputs[50]]))
+    outputs.append(Attention(use_scale=False, score_mode='concat')([inputs[49], inputs[50]]))
+
+    outputs.append(AdditiveAttention(use_scale=False)([inputs[49], inputs[50]]))
+    outputs.append(AdditiveAttention(use_scale=False)([inputs[49], inputs[50], inputs[51]]))
+    outputs.append(AdditiveAttention(use_scale=True)([inputs[49], inputs[50]]))
+    outputs.append(AdditiveAttention(use_scale=True)([inputs[49], inputs[50], inputs[51]]))
+
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=1, value_dim=None,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=1, value_dim=None,
+        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=2, value_dim=None,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=2, value_dim=None,
+        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=1, value_dim=2,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=1, value_dim=2,
+        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=3, key_dim=1, value_dim=None,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=3, key_dim=1, value_dim=None,
+        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
+    outputs.append(MultiHeadAttention(
+        num_heads=1, key_dim=1, value_dim=None,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))
+    outputs.append(MultiHeadAttention(
+        num_heads=2, key_dim=3, value_dim=5,
+        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))
+    outputs.append(MultiHeadAttention(
+        num_heads=2, key_dim=3, value_dim=5,
+        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))
 
     shared_conv = Conv2D(1, (1, 1),
                          padding='valid', name='shared_conv', activation='relu')
@@ -991,7 +1067,7 @@ def main():
 
         model_func = get_model_functions[model_name]
         model = model_func()
-        model.save(dest_path, include_optimizer=False)
+        model.save(dest_path)
 
         # Make sure models can be loaded again,
         # see https://github.com/fchollet/keras/issues/7682
